#ifndef MNIST_H
#define MNIST_H

#include <fstream>
#include <iomanip>
#include <tuple>

#include "opencv2/core.hpp"
#include <opencv2/opencv.hpp>

#include "aidge/data/Database.hpp"
#include "aidge/stimuli/Stimuli.hpp"
#include "aidge/graph/GraphView.hpp"
#include "aidge/scheduler/Scheduler.hpp"
#include "aidge/data/Tensor.hpp"
#include "aidge/backend/opencv/utils/Utils.hpp"

// #include "aidge/backend/opencv/data/TensorImpl.hpp"

namespace Aidge {

template <class T> void swapEndian(T& obj)
{
    unsigned char* memp = reinterpret_cast<unsigned char*>(&obj);
    std::reverse(memp, memp + sizeof(T));
}

inline bool isBigEndian()
{
    const union {
        uint32_t i;
        char c[4];
    } bint = {0x01020304};

    return bint.c[0] == 1;
}

class MNIST : public Database {
public:
    MNIST(const std::string& dataPath, 
            // const GraphView transformations, 
            bool train,
            bool loadDataInMemory = false)
    : Database(),
      mDataPath(dataPath),
    // mDataTransformations(transformations),
      mTrain(train),
      mLoadDataInMemory(loadDataInMemory)
    {
        // Uncompress train database 
        if (mTrain){
            MNIST::uncompress(mDataPath + "/train-images-idx3-ubyte",
                                dataPath + "/train-labels-idx1-ubyte");
        }else { // Uncompress test database 
            MNIST::uncompress(mDataPath + "/t10k-images-idx3-ubyte",
                                dataPath + "/t10k-labels-idx1-ubyte");
        }
    }
    
    std::vector<std::shared_ptr<Tensor>> getItem(std::size_t index) override;

    void uncompress(const std::string& dataPath,
                    const std::string& labelPath);

    std::size_t getLen() override;

    std::size_t getNbModalities() override;

    union MagicNumber {
        unsigned int value;
        unsigned char byte[4];
    };

    enum DataType {
        Unsigned = 0x08,
        Signed = 0x09,
        Short = 0x0B,
        Int = 0x0C,
        Float = 0x0D,
        Double = 0x0E
    }; 

protected:

    /// Stimuli data path
    std::string mDataPath;
    
    // True select the train database, False Select the test database
    bool mTrain;

    // True Load images in memory, False reload at each call
    bool mLoadDataInMemory;

    /// Stimuli data
    // Each index of the vector is one item of the database
    // One item of the MNIST database is the tuple <Image,label>
    // First stimuli of the tuple is a gray scale image stimuli of a writen digit
    // Second stimuli of the tuple is the label associated to the digit : unsigned integer 0-9
    std::vector<std::tuple<Stimuli,Stimuli>> mStimulis;

    /// Data Transformations
    // Data transformations use the GraphView mecanism
    // Transformations are a sequential graph with each operator of the graph being one transformation
    // GraphView mDataTransformations;

    // Scheduler to run the graph of data transformations
    // Scheduler mScheduler;

};
}

#endif // MNIST_H
