/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_BACKEND_OPENCV_UTILS_ATTRIBUTES_H_
#define AIDGE_BACKEND_OPENCV_UTILS_ATTRIBUTES_H_


#include <tuple>

#include "opencv2/core.hpp"
#include <opencv2/opencv.hpp>

#include "aidge/data/Tensor.hpp"
#include "aidge/backend/opencv/data/TensorImpl.hpp"
#include "aidge/backend/cpu/data/TensorImpl.hpp"
#include "aidge/data/Data.hpp"

namespace Aidge {

    /**
     * @brief Instanciate an aidge tensor with backend "opencv" from an opencv matrix
     * 
     * @param mat the cv::mat to instanciate the tensor from
     * @return std::shared_ptr<Tensor> aidge tensor
     */
    inline std::shared_ptr<Tensor> tensorOpencv(cv::Mat mat){
        // Get Mat dims
        std::vector<DimSize_t> matDims = std::vector<DimSize_t>({static_cast<DimSize_t>(mat.cols),
                                                                static_cast<DimSize_t>(mat.rows),
                                                                static_cast<DimSize_t>(mat.channels())});
        // Create tensor from the dims of the Cv::Mat
        std::shared_ptr<Tensor> tensor = std::make_shared<Tensor>(matDims);
        // Set beackend opencv 
        tensor->setBackend("opencv");
        // Set Data Type
        switch (mat.depth()) {
            case CV_8U:
                tensor->setDataType(Aidge::DataType::UInt8);
                break;
            case CV_8S:
                tensor->setDataType(Aidge::DataType::Int8);
                break;
            case CV_16U:
                tensor->setDataType(Aidge::DataType::UInt16);
                break;
            case CV_16S:
                tensor->setDataType(Aidge::DataType::Int16);
                break;
            case CV_32S:
                tensor->setDataType(Aidge::DataType::Int32);
                break;
            case CV_32F:
                tensor->setDataType(Aidge::DataType::Float32);
                break;
            case CV_64F:
                tensor->setDataType(Aidge::DataType::Float64);
                break;
            default:
                throw std::runtime_error(
                    "Cannot convert cv::Mat to Tensor: incompatible types.");
            }

        // Cast the tensorImpl to access setCvMat function
        TensorImpl_opencv_* tImpl_opencv = dynamic_cast<TensorImpl_opencv_*>(tensor->getImpl().get());
        tImpl_opencv->setCvMat(mat);
        return tensor;
    }

    /**
     * @brief Copy the data from a source 2D cv::mat to a destination pointer with an offset
     * 
     * @tparam CV_T The standard type corresponding to the opencv data type
     * @param mat opencv 2D mat to copy the data from
     * @param data destination pointer 
     * @param offset offset an the destination data pointer 
     */
    template <class CV_T>
    void convert(const cv::Mat& mat, void* data, size_t offset);

    
    /**
     * @brief Convert a tensor backend opencv into a tensor backend cpu
     * 
     * @param tensorOpencv tensor with backend opencv (contains a cv::mat)
     * @return std::shared_ptr<Tensor> tensor backend cpu (contains a std::vector)
     */
    std::shared_ptr<Tensor> convertCpu(std::shared_ptr<Aidge::Tensor> tensorOpencv);

    
}  // namespace

#endif // AIDGE_BACKEND_OPENCV_UTILS_ATTRIBUTES_H_